import logging
import random

import requests
from requests import RequestException

from bubuku.amazon import Amazon
from bubuku.zookeeper import AddressListProvider

_LOG = logging.getLogger('bubuku.zookeeper.exhibitor')


class AWSExhibitorAddressProvider(AddressListProvider):
    def __init__(self, amazon: Amazon, zk_stack_name: str):
        self.amazon = amazon
        self.zk_stack_name = zk_stack_name
        self.exhibitors = []

    def get_latest_address(self) -> (list, int):
        json_ = self._query_exhibitors(self.exhibitors)
        if not json_:
            self.exhibitors = self.query_from_amazon()
            json_ = self._query_exhibitors(self.exhibitors)
        if isinstance(json_, dict) and 'servers' in json_ and 'port' in json_:
            self.exhibitors = json_['servers']
            return json_['servers'], int(json_['port'])
        return None

    def _query_exhibitors(self, exhibitors):
        if not exhibitors:
            return None
        random.shuffle(exhibitors)
        for host in exhibitors:
            url = 'http://{}:{}{}'.format(host, 8181, '/exhibitor/v1/cluster/list')
            try:
                response = requests.get(url, timeout=3.1)
                return response.json()
            except RequestException as e:
                _LOG.warn('Failed to query zookeeper list information from {}'.format(url), exc_info=e)
        return None

    def query_from_amazon(self):
        return self.amazon.get_addresses_by_lb_name(self.zk_stack_name)



import subprocess


class CmdHelper(object):
    def get_disk_stats(self) -> (int, int):
        """
        Returns total disk stats,
        :return: used_kb, free_kb
        """
        disks = self.cmd_run("df -k | tail -n +2 |  awk '{ print $3, $4 }'").split("\n")
        total_used = total_free = 0
        for disk in disks:
            parts = disk.split(" ")
            if len(parts) == 2:
                used, free = tuple(parts)
                total_used += int(used)
                total_free += int(free)
        return total_used, total_free

    def cmd_run(self, cmd):
        output = subprocess.check_output(cmd, shell=True)
        return output.decode("utf-8")



import json
import logging
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer

from bubuku.utils import CmdHelper

_LOG = logging.getLogger('bubuku.health')


class _Handler(BaseHTTPRequestHandler):
    cmd_helper = None

    def do_GET(self):
        if self.path in ('/api/disk_stats', '/api/disk_stats/'):
            used_kb, free_kb = self.cmd_helper.get_disk_stats()
            self._send_response({'free_kb': free_kb, 'used_kb': used_kb})
        else:
            self._send_response({'status': 'OK'})

    def _send_response(self, json_):
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(json_).encode('utf-8'))


def start_server(port, cmd_helper: CmdHelper) -> threading.Thread:
    def _thread_func():
        _Handler.cmd_helper = cmd_helper
        server = HTTPServer(('', port), _Handler)
        server.serve_forever()
        server.socket.close()

    t = threading.Thread(target=_thread_func, daemon=True)
    _LOG.info('Starting health server on port {}'.format(port))
    t.start()
    return t



import logging

from bubuku.broker import BrokerManager
from bubuku.controller import Check, Change
from bubuku.features.migrate import MigrationChange
from bubuku.features.rebalance import RebalanceChange
from bubuku.features.restart_on_zk_change import RestartBrokerChange
from bubuku.features.swap_partitions import SwapPartitionsChange, load_swap_data
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.features.remote_exec')


class RemoteCommandExecutorCheck(Check):
    def __init__(self, zk: BukuExhibitor, broker_manager: BrokerManager, api_port):
        super().__init__(check_interval_s=30)
        self.zk = zk
        self.broker_manager = broker_manager
        self.api_port = api_port

    def check(self) -> Change:
        with self.zk.lock():
            data = self.zk.take_action(self.broker_manager.id_manager.get_broker_id())
        if not data:
            return None
        if 'name' not in data:
            _LOG.error('Action name can not be restored from {}, skipping'.format(data))
            return None
        try:
            if data['name'] == 'restart':
                return RestartBrokerChange(self.zk, self.broker_manager, lambda: False)
            elif data['name'] == 'rebalance':
                return RebalanceChange(self.zk, self.zk.get_broker_ids())
            elif data['name'] == 'migrate':
                return MigrationChange(self.zk, data['from'], data['to'], data['shrink'])
            elif data['name'] == 'fatboyslim':
                return SwapPartitionsChange(self.zk,
                                            lambda x: load_swap_data(x, self.api_port, int(data['threshold_kb'])))
            else:
                _LOG.error('Action {} not supported'.format(data))
        except Exception as e:
            _LOG.error('Failed to create action from {}'.format(data), exc_info=e)
        return None

    def __str__(self):
        return 'RemoteCommandExecutorCheck'

    @staticmethod
    def register_restart(zk: BukuExhibitor, broker_id: str):
        with zk.lock():
            zk.register_action(
                {'name': 'restart'},
                broker_id=broker_id)

    @staticmethod
    def register_rebalance(zk: BukuExhibitor, broker_id: str):
        with zk.lock():
            if broker_id:
                zk.register_action({'name': 'rebalance'}, broker_id=broker_id)
            else:
                zk.register_action({'name': 'rebalance'})

    @staticmethod
    def register_migration(zk: BukuExhibitor, brokers_from: list, brokers_to: list, shrink: bool, broker_id: str):
        if len(brokers_from) != len(brokers_to):
            raise Exception('Brokers list {} and {} must have the same size'.format(brokers_from, brokers_to))
        if any(b in brokers_from for b in brokers_to) or any(b in brokers_to for b in brokers_from):
            raise Exception('Broker lists can not hold same broker ids')

        if len(set(brokers_from)) != len(brokers_from):
            raise Exception('Can not use same broker ids for source_list {}'.format(brokers_from))
        if len(set(brokers_to)) != len(brokers_to):
            raise Exception('Can not use same broker ids for source_list {}'.format(brokers_from))

        active_ids = zk.get_broker_ids()
        if any(b not in active_ids for b in brokers_from) or any(b not in active_ids for b in brokers_to):
            raise Exception('Brokers dead from: {} to: {} alive:{}'.format(brokers_from, brokers_to, active_ids))

        if broker_id and str(broker_id) not in active_ids:
            raise Exception('Broker id to run change on ({}) is not in active list {}'.format(
                broker_id, active_ids))

        with zk.lock():
            action = {'name': 'migrate', 'from': brokers_from, 'to': brokers_to, 'shrink': bool(shrink)}
            if broker_id:
                zk.register_action(action, str(broker_id))
            else:
                zk.register_action(action)

    @staticmethod
    def register_fatboy_slim(zk: BukuExhibitor, threshold_kb: int):
        if zk.is_rebalancing():
            _LOG.warn('Rebalance is already in progress, may be it will take time for this command to start processing')
        with zk.lock():
            zk.register_action({'name': 'fatboyslim', 'threshold_kb': threshold_kb})
